
/**
 * @file menu.h
 * @copyright 2013, 2014 www.enkore.de
 *
 * Modular menu system for embedded applications
 */

#include <stdint.h>

/**
 * Number of menu items displayed at once
 */
#define MENU_PAGE_ITEMS 4

/**
 * Possible user inputs for menus
 */
typedef enum {
    MENU_INPUT_ENTER,
    MENU_INPUT_NEXT,
    MENU_INPUT_PREV,
    MENU_INPUT_CANCEL
} menu_input;

/**
 * Menu entry types
 */
typedef enum {
    MENU_TYPE_NORMAL,
    MENU_TYPE_CALLBACK,
} menu_type;

/**
 * Flags for displaying entries
 */
typedef enum {
    /// Entry just sits there.
    MENU_DFLAG_NONE = 0,
    /// The entry is selected
    MENU_DFLAG_SELECTED = 0x01,
} menu_display_flags;

typedef struct menu_entry_t menu_entry_t;

/**
 * Menu entry callback
 *
 * @param data the data member of the selected menu_entry_t
 * @returns false to close menu, true to return to menu
 * @see menu_entry_t
 */
typedef int (*menu_entry_callback_t)(void *data);

/**
 * Menu display function
 *
 * @param line Display line
 * @param display_flags @see menu_display_flags
 * @param entry menu entry to display
 */
typedef void (*menu_write_t)(uint8_t line, uint8_t display_flags, const menu_entry_t *entry); 

/**
 * Called right before the first batched calls to menu_write_t
 */
typedef void (*menu_write_begin_t)(void);

/**
 * Struct for menu entries
 */
struct menu_entry_t {
    /// Displayed text
    const char *text;
    /// Entry type
    const menu_type type;
    /// Entry data valid fields depend on type
    union {
	/// Data for MENU_TYPE_CALLBACK
	struct {
	    menu_entry_callback_t cb;
	    void *data;
	} callback;
    } data;
};

/**
 * Struct for a complete menu, containing all necessary data and state.
 */
typedef struct {
    /// Array of entries
    menu_entry_t *entries;
    /// Number of entries
    int num_entries;
    /// Selected entry (0-indexed)
    int selection;

    /// Writer function
    menu_write_t write;
    /// Writer start function
    menu_write_begin_t write_begin;
} menu_t;

/**
 * Enter menu, i.e.
 *
 * * Set selection to beginning of menu
 * * Display first page of menu
 *
 * @param menu menu
 */
void menu_state_init(menu_t *menu);

/**
 * Process users input
 *
 * @param menu initialized menu
 * @param input valid user input
 * @pre menu_state_init(menu)
 */
int menu_state_advance(menu_t *menu, menu_input input);

/**
 * Display a menu and return selection
 *
 * @param menu menu
 * @param input user input function (blocking)
 * @returns selected menu item or -1 on cancel
 */
int menu_display(menu_t *menu, menu_input (*input)(void));
